# noqa
# pragma: no cover
from biolib.pyppeteer.pyppeteer.models import Devices

devices: Devices = {
    "Blackberry PlayBook": {
        "userAgent": "Mozilla/5.0 (PlayBook; U; RIM Tablet OS 2.1.0; en-US) AppleWebKit/536.2+ (KHTML like Gecko) Version/7.2.1.0 Safari/536.2+",
        "viewport": {
            "width": 600,
            "height": 1024,
            "deviceScaleFactor": 1,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Blackberry PlayBook landscape": {
        "userAgent": "Mozilla/5.0 (PlayBook; U; RIM Tablet OS 2.1.0; en-US) AppleWebKit/536.2+ (KHTML like Gecko) Version/7.2.1.0 Safari/536.2+",
        "viewport": {
            "width": 1024,
            "height": 600,
            "deviceScaleFactor": 1,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "BlackBerry Z30": {
        "userAgent": "Mozilla/5.0 (BB10; Touch) AppleWebKit/537.10+ (KHTML, like Gecko) Version/10.0.9.2372 Mobile Safari/537.10+",
        "viewport": {
            "width": 360,
            "height": 640,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "BlackBerry Z30 landscape": {
        "userAgent": "Mozilla/5.0 (BB10; Touch) AppleWebKit/537.10+ (KHTML, like Gecko) Version/10.0.9.2372 Mobile Safari/537.10+",
        "viewport": {
            "width": 640,
            "height": 360,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Galaxy Note 3": {
        "userAgent": "Mozilla/5.0 (Linux; U; Android 4.3; en-us; SM-N900T Build/JSS15J) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30",
        "viewport": {
            "width": 360,
            "height": 640,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Galaxy Note 3 landscape": {
        "userAgent": "Mozilla/5.0 (Linux; U; Android 4.3; en-us; SM-N900T Build/JSS15J) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30",
        "viewport": {
            "width": 640,
            "height": 360,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Galaxy Note II": {
        "userAgent": "Mozilla/5.0 (Linux; U; Android 4.1; en-us; GT-N7100 Build/JRO03C) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30",
        "viewport": {
            "width": 360,
            "height": 640,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Galaxy Note II landscape": {
        "userAgent": "Mozilla/5.0 (Linux; U; Android 4.1; en-us; GT-N7100 Build/JRO03C) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30",
        "viewport": {
            "width": 640,
            "height": 360,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Galaxy S III": {
        "userAgent": "Mozilla/5.0 (Linux; U; Android 4.0; en-us; GT-I9300 Build/IMM76D) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30",
        "viewport": {
            "width": 360,
            "height": 640,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Galaxy S III landscape": {
        "userAgent": "Mozilla/5.0 (Linux; U; Android 4.0; en-us; GT-I9300 Build/IMM76D) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30",
        "viewport": {
            "width": 640,
            "height": 360,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Galaxy S5": {
        "userAgent": "Mozilla/5.0 (Linux; Android 5.0; SM-G900P Build/LRX21T) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 360,
            "height": 640,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Galaxy S5 landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 5.0; SM-G900P Build/LRX21T) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 640,
            "height": 360,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPad": {
        "userAgent": "Mozilla/5.0 (iPad; CPU OS 11_0 like Mac OS X) AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0 Mobile/15A5341f Safari/604.1",
        "viewport": {
            "width": 768,
            "height": 1024,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPad landscape": {
        "userAgent": "Mozilla/5.0 (iPad; CPU OS 11_0 like Mac OS X) AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0 Mobile/15A5341f Safari/604.1",
        "viewport": {
            "width": 1024,
            "height": 768,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPad Mini": {
        "userAgent": "Mozilla/5.0 (iPad; CPU OS 11_0 like Mac OS X) AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0 Mobile/15A5341f Safari/604.1",
        "viewport": {
            "width": 768,
            "height": 1024,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPad Mini landscape": {
        "userAgent": "Mozilla/5.0 (iPad; CPU OS 11_0 like Mac OS X) AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0 Mobile/15A5341f Safari/604.1",
        "viewport": {
            "width": 1024,
            "height": 768,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPad Pro": {
        "userAgent": "Mozilla/5.0 (iPad; CPU OS 11_0 like Mac OS X) AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0 Mobile/15A5341f Safari/604.1",
        "viewport": {
            "width": 1024,
            "height": 1366,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPad Pro landscape": {
        "userAgent": "Mozilla/5.0 (iPad; CPU OS 11_0 like Mac OS X) AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0 Mobile/15A5341f Safari/604.1",
        "viewport": {
            "width": 1366,
            "height": 1024,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone 4": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 7_1_2 like Mac OS X) AppleWebKit/537.51.2 (KHTML, like Gecko) Version/7.0 Mobile/11D257 Safari/9537.53",
        "viewport": {
            "width": 320,
            "height": 480,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone 4 landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 7_1_2 like Mac OS X) AppleWebKit/537.51.2 (KHTML, like Gecko) Version/7.0 Mobile/11D257 Safari/9537.53",
        "viewport": {
            "width": 480,
            "height": 320,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone 5": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3_1 like Mac OS X) AppleWebKit/603.1.30 (KHTML, like Gecko) Version/10.0 Mobile/14E304 Safari/602.1",
        "viewport": {
            "width": 320,
            "height": 568,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone 5 landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3_1 like Mac OS X) AppleWebKit/603.1.30 (KHTML, like Gecko) Version/10.0 Mobile/14E304 Safari/602.1",
        "viewport": {
            "width": 568,
            "height": 320,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone 6": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 375,
            "height": 667,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone 6 landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 667,
            "height": 375,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone 6 Plus": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 414,
            "height": 736,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone 6 Plus landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 736,
            "height": 414,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone 7": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 375,
            "height": 667,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone 7 landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 667,
            "height": 375,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone 7 Plus": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 414,
            "height": 736,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone 7 Plus landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 736,
            "height": 414,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone 8": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 375,
            "height": 667,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone 8 landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 667,
            "height": 375,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone 8 Plus": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 414,
            "height": 736,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone 8 Plus landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 736,
            "height": 414,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone SE": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3_1 like Mac OS X) AppleWebKit/603.1.30 (KHTML, like Gecko) Version/10.0 Mobile/14E304 Safari/602.1",
        "viewport": {
            "width": 320,
            "height": 568,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone SE landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3_1 like Mac OS X) AppleWebKit/603.1.30 (KHTML, like Gecko) Version/10.0 Mobile/14E304 Safari/602.1",
        "viewport": {
            "width": 568,
            "height": 320,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone X": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 375,
            "height": 812,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone X landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "viewport": {
            "width": 812,
            "height": 375,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "iPhone XR": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1",
        "viewport": {
            "width": 414,
            "height": 896,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "iPhone XR landscape": {
        "userAgent": "Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1",
        "viewport": {
            "width": 896,
            "height": 414,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "JioPhone 2": {
        "userAgent": "Mozilla/5.0 (Mobile; LYF/F300B/LYF-F300B-001-01-15-130718-i;Android; rv:48.0) Gecko/48.0 Firefox/48.0 KAIOS/2.5",
        "viewport": {
            "width": 240,
            "height": 320,
            "deviceScaleFactor": 1,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "JioPhone 2 landscape": {
        "userAgent": "Mozilla/5.0 (Mobile; LYF/F300B/LYF-F300B-001-01-15-130718-i;Android; rv:48.0) Gecko/48.0 Firefox/48.0 KAIOS/2.5",
        "viewport": {
            "width": 320,
            "height": 240,
            "deviceScaleFactor": 1,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Kindle Fire HDX": {
        "userAgent": "Mozilla/5.0 (Linux; U; en-us; KFAPWI Build/JDQ39) AppleWebKit/535.19 (KHTML, like Gecko) Silk/3.13 Safari/535.19 Silk-Accelerated=True",
        "viewport": {
            "width": 800,
            "height": 1280,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Kindle Fire HDX landscape": {
        "userAgent": "Mozilla/5.0 (Linux; U; en-us; KFAPWI Build/JDQ39) AppleWebKit/535.19 (KHTML, like Gecko) Silk/3.13 Safari/535.19 Silk-Accelerated=True",
        "viewport": {
            "width": 1280,
            "height": 800,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "LG Optimus L70": {
        "userAgent": "Mozilla/5.0 (Linux; U; Android 4.4.2; en-us; LGMS323 Build/KOT49I.MS32310c) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 384,
            "height": 640,
            "deviceScaleFactor": 1.25,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "LG Optimus L70 landscape": {
        "userAgent": "Mozilla/5.0 (Linux; U; Android 4.4.2; en-us; LGMS323 Build/KOT49I.MS32310c) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 640,
            "height": 384,
            "deviceScaleFactor": 1.25,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Microsoft Lumia 550": {
        "userAgent": "Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Microsoft; Lumia 550) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0 Mobile Safari/537.36 Edge/14.14263",
        "viewport": {
            "width": 640,
            "height": 360,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Microsoft Lumia 950": {
        "userAgent": "Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Microsoft; Lumia 950) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0 Mobile Safari/537.36 Edge/14.14263",
        "viewport": {
            "width": 360,
            "height": 640,
            "deviceScaleFactor": 4,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Microsoft Lumia 950 landscape": {
        "userAgent": "Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Microsoft; Lumia 950) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0 Mobile Safari/537.36 Edge/14.14263",
        "viewport": {
            "width": 640,
            "height": 360,
            "deviceScaleFactor": 4,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Nexus 10": {
        "userAgent": "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 10 Build/MOB31T) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Safari/537.36",
        "viewport": {
            "width": 800,
            "height": 1280,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Nexus 10 landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 10 Build/MOB31T) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Safari/537.36",
        "viewport": {
            "width": 1280,
            "height": 800,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Nexus 4": {
        "userAgent": "Mozilla/5.0 (Linux; Android 4.4.2; Nexus 4 Build/KOT49H) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 384,
            "height": 640,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Nexus 4 landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 4.4.2; Nexus 4 Build/KOT49H) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 640,
            "height": 384,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Nexus 5": {
        "userAgent": "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 360,
            "height": 640,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Nexus 5 landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 640,
            "height": 360,
            "deviceScaleFactor": 3,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Nexus 5X": {
        "userAgent": "Mozilla/5.0 (Linux; Android 8.0.0; Nexus 5X Build/OPR4.170623.006) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 412,
            "height": 732,
            "deviceScaleFactor": 2.625,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Nexus 5X landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 8.0.0; Nexus 5X Build/OPR4.170623.006) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 732,
            "height": 412,
            "deviceScaleFactor": 2.625,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Nexus 6": {
        "userAgent": "Mozilla/5.0 (Linux; Android 7.1.1; Nexus 6 Build/N6F26U) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 412,
            "height": 732,
            "deviceScaleFactor": 3.5,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Nexus 6 landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 7.1.1; Nexus 6 Build/N6F26U) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 732,
            "height": 412,
            "deviceScaleFactor": 3.5,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Nexus 6P": {
        "userAgent": "Mozilla/5.0 (Linux; Android 8.0.0; Nexus 6P Build/OPP3.170518.006) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 412,
            "height": 732,
            "deviceScaleFactor": 3.5,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Nexus 6P landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 8.0.0; Nexus 6P Build/OPP3.170518.006) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 732,
            "height": 412,
            "deviceScaleFactor": 3.5,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Nexus 7": {
        "userAgent": "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 7 Build/MOB30X) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Safari/537.36",
        "viewport": {
            "width": 600,
            "height": 960,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Nexus 7 landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 7 Build/MOB30X) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Safari/537.36",
        "viewport": {
            "width": 960,
            "height": 600,
            "deviceScaleFactor": 2,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Nokia Lumia 520": {
        "userAgent": "Mozilla/5.0 (compatible; MSIE 10.0; Windows Phone 8.0; Trident/6.0; IEMobile/10.0; ARM; Touch; NOKIA; Lumia 520)",
        "viewport": {
            "width": 320,
            "height": 533,
            "deviceScaleFactor": 1.5,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Nokia Lumia 520 landscape": {
        "userAgent": "Mozilla/5.0 (compatible; MSIE 10.0; Windows Phone 8.0; Trident/6.0; IEMobile/10.0; ARM; Touch; NOKIA; Lumia 520)",
        "viewport": {
            "width": 533,
            "height": 320,
            "deviceScaleFactor": 1.5,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Nokia N9": {
        "userAgent": "Mozilla/5.0 (MeeGo; NokiaN9) AppleWebKit/534.13 (KHTML, like Gecko) NokiaBrowser/8.5.0 Mobile Safari/534.13",
        "viewport": {
            "width": 480,
            "height": 854,
            "deviceScaleFactor": 1,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Nokia N9 landscape": {
        "userAgent": "Mozilla/5.0 (MeeGo; NokiaN9) AppleWebKit/534.13 (KHTML, like Gecko) NokiaBrowser/8.5.0 Mobile Safari/534.13",
        "viewport": {
            "width": 854,
            "height": 480,
            "deviceScaleFactor": 1,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Pixel 2": {
        "userAgent": "Mozilla/5.0 (Linux; Android 8.0; Pixel 2 Build/OPD3.170816.012) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 411,
            "height": 731,
            "deviceScaleFactor": 2.625,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Pixel 2 landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 8.0; Pixel 2 Build/OPD3.170816.012) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 731,
            "height": 411,
            "deviceScaleFactor": 2.625,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
    "Pixel 2 XL": {
        "userAgent": "Mozilla/5.0 (Linux; Android 8.0.0; Pixel 2 XL Build/OPD1.170816.004) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 411,
            "height": 823,
            "deviceScaleFactor": 3.5,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": False,
        },
    },
    "Pixel 2 XL landscape": {
        "userAgent": "Mozilla/5.0 (Linux; Android 8.0.0; Pixel 2 XL Build/OPD1.170816.004) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3765.0 Mobile Safari/537.36",
        "viewport": {
            "width": 823,
            "height": 411,
            "deviceScaleFactor": 3.5,
            "isMobile": True,
            "hasTouch": True,
            "isLandscape": True,
        },
    },
}



from Crypto.Cipher import PKCS1_OAEP, AES
from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes
from Crypto.Hash import SHA256

from biolib.biolib_binary_format.base_bbf_package import BioLibBinaryFormatBasePackage


class RsaEncryptedAesPackage(BioLibBinaryFormatBasePackage):
    def __init__(self, bbf=None):
        super().__init__(bbf)
        self.package_type = 6
        self.iv_len = 12
        self.aes_key_len = 32
        self.tag_len = 16

    def create(self, rsa_public_key_der, unencrypted_bytes):
        public_key = RSA.import_key(rsa_public_key_der)
        rsa_cipher = PKCS1_OAEP.new(public_key, hashAlgo=SHA256)

        iv = get_random_bytes(self.iv_len)
        aes_key_buffer = get_random_bytes(self.aes_key_len)
        aes_key = AES.new(aes_key_buffer, AES.MODE_GCM, iv)
        rsa_encrypted_aes_key_buffer = rsa_cipher.encrypt(aes_key_buffer)
        aes_encrypted_data, tag = aes_key.encrypt_and_digest(bytes(unencrypted_bytes))

        return self.serialize(rsa_encrypted_aes_key_buffer, iv, tag, aes_encrypted_data), aes_key_buffer

    def serialize(self, encrypted_aes_key, iv, tag, encrypted_data):
        bbf_data = bytearray()
        bbf_data.extend(self.version.to_bytes(1, 'big'))
        bbf_data.extend(self.package_type.to_bytes(1, 'big'))

        bbf_data.extend(len(encrypted_aes_key).to_bytes(2, 'big'))
        bbf_data.extend(encrypted_aes_key)
        bbf_data.extend(iv)
        bbf_data.extend(tag)

        bbf_data.extend(len(encrypted_data).to_bytes(8, 'big'))
        bbf_data.extend(encrypted_data)

        return bbf_data

    def deserialize(self):
        version = self.get_data(1, output_type='int')
        package_type = self.get_data(1, output_type='int')
        self.check_version_and_type(version=version, package_type=package_type, expected_package_type=self.package_type)

        encrypted_aes_key_len = self.get_data(2, output_type='int')
        encrypted_aes_key = self.get_data(encrypted_aes_key_len)
        iv = self.get_data(self.iv_len)
        tag = self.get_data(self.tag_len)

        encrypted_data_len = self.get_data(8, output_type='int')
        encrypted_data = self.get_data(encrypted_data_len)

        return encrypted_aes_key, iv, tag, encrypted_data



import base64

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes

from biolib.biolib_binary_format import BioLibBinaryFormatBasePackage
from biolib.biolib_binary_format.base_package import IndexableBuffer, InMemoryIndexableBuffer
from biolib.biolib_binary_format.common_types import FilesInfo
from biolib.biolib_binary_format.unencrypted_module_output import unencrypted_module_output_metadata_lengths
from biolib.biolib_binary_format.module_output import ModuleOutput
from biolib.typing_utils import Optional, TypedDict, List

IV_LENGTH_BYTES = 12
TAG_LENGTH_BYTES = 16


class EncryptedModuleOutputMetadata(TypedDict):
    version: int
    type: int
    stdout_length: int
    stderr_length: int
    files_info_length: int
    files_data_length: int


encrypted_module_output_metadata_lengths = EncryptedModuleOutputMetadata(
    **unencrypted_module_output_metadata_lengths,  # type: ignore
)


class EncryptedModuleOutputError(BaseException):
    pass


class EncryptedModuleOutputWithKey(BioLibBinaryFormatBasePackage):
    _TYPE = 12
    _VERSION = 1

    def __init__(self, aes_key_string_b64: str, buffer: Optional[IndexableBuffer] = None):
        super().__init__()
        self._aes_key_buffer: bytes = base64.urlsafe_b64decode(aes_key_string_b64.encode())
        self._buffer: Optional[IndexableBuffer] = buffer

    def _get_metadata(self) -> EncryptedModuleOutputMetadata:
        if not self._buffer:
            raise EncryptedModuleOutputError(
                'You must construct this class with a buffer to get its metadata'
            )

        metadata = {}
        for field_name, field_length in encrypted_module_output_metadata_lengths.items():
            value = self._buffer.get_data_with_pointer_as_int(length_bytes=field_length)
            if field_name == 'version' and value != EncryptedModuleOutputWithKey._VERSION:
                raise Exception('Version does not match')

            if field_name == 'type' and value != EncryptedModuleOutputWithKey._TYPE:
                raise Exception('Type does not match')

            metadata[field_name] = value

        return EncryptedModuleOutputMetadata(**metadata)  # type: ignore

    @property
    def _metadata_length_bytes(self) -> int:
        return sum(encrypted_module_output_metadata_lengths.values())

    def convert_to_serialized_module_output(self) -> bytes:
        if not self._buffer:
            raise EncryptedModuleOutputError('You must construct this class with a buffer to convert to module output')

        metadata = self._get_metadata()
        exit_code_bytes = self._decrypt(
            self._buffer.get_data_with_pointer(length_bytes=2 + IV_LENGTH_BYTES + TAG_LENGTH_BYTES)
        )
        exit_code = int.from_bytes(exit_code_bytes, 'big')

        stdout = self._decrypt(
            self._buffer.get_data_with_pointer(length_bytes=metadata['stdout_length'])
        )

        stderr = self._decrypt(
            self._buffer.get_data_with_pointer(length_bytes=metadata['stderr_length'])
        )

        files_info_serialized = InMemoryIndexableBuffer(self._decrypt(
            self._buffer.get_data_with_pointer(length_bytes=metadata['files_info_length'])
        ))

        files_info: List[FilesInfo] = []
        while files_info_serialized.pointer <= len(files_info_serialized) - 1:  # subtract 1 as pointer is 0-indexed
            path_length = files_info_serialized.get_data_with_pointer_as_int(length_bytes=4)
            files_info.append({
                'path': files_info_serialized.get_data_with_pointer(length_bytes=path_length).decode(),
                'data_length': files_info_serialized.get_data_with_pointer_as_int(length_bytes=8)
            })

        files = {}
        for file_info in files_info:
            files[file_info['path']] = self._decrypt(
                self._buffer.get_data_with_pointer(length_bytes=file_info['data_length'])
            )

        return ModuleOutput().serialize(
            stdout=stdout,
            stderr=stderr,
            exit_code=exit_code,
            files=files
        )

    def create_from_serialized_module_output(self, module_output_serialized: bytes) -> bytes:
        module_output_dict = ModuleOutput(module_output_serialized).deserialize()

        bbf_data = bytearray()
        bbf_data.extend(EncryptedModuleOutputWithKey._VERSION.to_bytes(1, 'big'))
        bbf_data.extend(EncryptedModuleOutputWithKey._TYPE.to_bytes(1, 'big'))

        encrypted_stdout = self._encrypt(module_output_dict['stdout'])
        encrypted_stderr = self._encrypt(module_output_dict['stderr'])

        files_info = bytearray()
        encrypted_files_data = bytearray()
        for path, data in module_output_dict['files'].items():
            encoded_path = path.encode()
            files_info.extend(len(encoded_path).to_bytes(4, 'big'))
            files_info.extend(encoded_path)

            encrypted_data = self._encrypt(data)
            files_info.extend(len(encrypted_data).to_bytes(8, 'big'))
            encrypted_files_data.extend(encrypted_data)

        encrypted_files_info = self._encrypt(files_info)

        # Length of encrypted stdout and stderr
        bbf_data.extend(len(encrypted_stdout).to_bytes(8, 'big'))
        bbf_data.extend(len(encrypted_stderr).to_bytes(8, 'big'))

        bbf_data.extend(len(encrypted_files_info).to_bytes(8, 'big'))
        bbf_data.extend(len(encrypted_files_data).to_bytes(8, 'big'))

        bbf_data.extend(self._encrypt(
            module_output_dict['exit_code'].to_bytes(2, 'big')
        ))
        bbf_data.extend(encrypted_stdout)
        bbf_data.extend(encrypted_stderr)

        bbf_data.extend(encrypted_files_info)
        bbf_data.extend(encrypted_files_data)

        return bbf_data

    def _encrypt(self, data: bytes) -> bytes:
        iv = get_random_bytes(12)
        aes_key = AES.new(self._aes_key_buffer, AES.MODE_GCM, iv)
        encrypted_data, tag = aes_key.encrypt_and_digest(data)  # type: ignore
        return iv + encrypted_data + tag

    def _decrypt(self, data: bytes) -> bytes:
        iv = data[0:IV_LENGTH_BYTES]
        tag = data[-TAG_LENGTH_BYTES:]
        data = data[IV_LENGTH_BYTES:-TAG_LENGTH_BYTES]
        cipher = AES.new(self._aes_key_buffer, AES.MODE_GCM, iv)
        plaintext = cipher.decrypt(data)
        try:
            cipher.verify(tag)  # type: ignore
            return plaintext
        except ValueError as error:
            raise EncryptedModuleOutputError('Decryption failed: Key incorrect or message corrupted') from error



import base64

import requests
from Crypto.Random import get_random_bytes

from biolib import utils
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_binary_format.base_package import InMemoryIndexableBuffer
from biolib.biolib_binary_format.encrypted_module_output import EncryptedModuleOutputWithKey
from biolib.biolib_binary_format.unencrypted_module_output import UnencryptedModuleOutput
from biolib.biolib_errors import BioLibError
from biolib.compute_node.job_worker.job_key_cache import JobKeyCacheState
from biolib.biolib_logging import logger_no_user_data, logger


class JobStorage:

    @staticmethod
    def upload_module_output(job_id: str, module_output: bytes, aes_key_string_b64: str):
        try:
            if utils.DISABLE_CLIENT_SIDE_ENCRYPTION:
                storage_module_output = UnencryptedModuleOutput.create_from_serialized_module_output(module_output)
                logger_no_user_data.debug(f'Job "{job_id}" uploading result to S3...')

            else:
                if not aes_key_string_b64:
                    raise Exception('Missing AES key for module output upload as client side encryption is enabled')
                storage_module_output = EncryptedModuleOutputWithKey(
                    aes_key_string_b64=aes_key_string_b64
                ).create_from_serialized_module_output(module_output)
                logger_no_user_data.debug(f'Job "{job_id}" uploading encrypted result to S3...')

        except Exception as error:
            logger_no_user_data.debug('Failed to get storage module output from serialized module output')
            logger.debug(f'Failed to get storage module output from serialized module output due to {error}')
            raise error

        try:
            presigned_upload_link = BiolibJobApi.get_job_storage_result_upload_url(job_id)

            request = requests.put(
                data=storage_module_output,
                url=presigned_upload_link,
                timeout=3600,  # timeout after 1 hour
            )

            if not request.ok:
                raise BioLibError('Failed to upload result to S3')

        except Exception as error:
            logger_no_user_data.debug('Failed to upload result to S3')
            logger.debug(f"Failed to upload result to S3 due to {error}")
            raise error

    @staticmethod
    def get_result(job):
        try:
            presigned_download_url = BiolibJobApi.get_job_storage_result_download_url(job['auth_token'])
            result_response = requests.get(
                url=presigned_download_url,
                timeout=3600,  # timeout after 1 hour
            )

            if not result_response.ok:
                raise BioLibError(result_response.content)

            storage_module_output = result_response.content

        except Exception as error:
            logger.debug(f'Failed to get results from S3 due to {error}')
            raise error

        if utils.BASE_URL_IS_PUBLIC_BIOLIB:
            with JobKeyCacheState() as cache_state:
                aes_key_string_b64 = cache_state[job['public_id']]

            return EncryptedModuleOutputWithKey(
                buffer=InMemoryIndexableBuffer(storage_module_output),
                aes_key_string_b64=aes_key_string_b64
            ).convert_to_serialized_module_output()

        else:
            return UnencryptedModuleOutput(
                buffer=InMemoryIndexableBuffer(storage_module_output)
            ).convert_to_serialized_module_output()

    @staticmethod
    def generate_and_store_key_buffer_for_job(job_id):
        aes_key_buffer = get_random_bytes(32)
        aes_key_string_b64 = base64.urlsafe_b64encode(aes_key_buffer).decode()

        with JobKeyCacheState() as cache_state:
            cache_state[job_id] = aes_key_string_b64

        return aes_key_string_b64



from typing import TYPE_CHECKING, Dict, List, Set, Union

if TYPE_CHECKING:
    from pyppeteer.connection.cdpsession import CDPSession
    from pyppeteer.jshandle import ElementHandle


class Accessibility:
    def __init__(self, client: 'CDPSession'):
        self._client = client

    async def snapshot(self, interestingOnly: bool = True, root: 'ElementHandle' = None):
        nodes = (await self._client.send('Accessibility.getFullAXTree'))['nodes']
        backendNodeId = None
        if root:
            node = (await self._client.send('DOM.describeNode', {'objectId': root._remoteObject['objectId']}))['node']
            backendNodeId = node['backendNodeId']
        defaultRoot = AXNode.createTree(nodes)
        needle = defaultRoot
        if backendNodeId:
            needle = defaultRoot.find(lambda _node: _node._payload['backendDOMNodeId'] == backendNodeId)
            if not needle:
                return
        if not interestingOnly:
            return serializeTree(needle)[0]

        interestingNodes = set()
        collectInterestingNodes(interestingNodes, defaultRoot, False)
        if needle not in interestingNodes:
            return None
        return serializeTree(needle, interestingNodes)[0]


class AXNode:
    def __init__(self, payload: Dict):
        """
        :param payload: dict with keys nodeId, name, role, properties
        """
        self._payload = payload
        self._children: List[AXNode] = []
        self._richlyEditable = False
        self._editable = False
        self._focusable = False
        self._expanded = False
        self._hidden = False
        self._name = payload.get('name', {}).get('value', '')
        self._role = payload.get('role', {}).get('value', 'Unknown')
        self._cacheHasFocusableChild = None

        for property in payload.get('properties', []):
            _name = property['name']
            _value = property.get('value', {}).get('value', None)
            if _name == 'editable':
                self._richlyEditable = _value == 'richtext'
                self._editable = True
            if _name == 'focusable':
                self._focusable = _value
            if _name == 'expanded':
                self._expanded = _value
            if _name == 'hidden':
                self._hidden = _value

    @property
    def _isPlainTextField(self):
        if self._richlyEditable:
            return False
        if self._editable:
            return True
        return self._role in ['textbox', 'ComboBox', 'searchbox']

    @property
    def _isTextOnlyObject(self):
        return self._role in ['LineBreak', 'text', 'InlineTextBox']

    @property
    def _hasFocusableChild(self):
        if self._cacheHasFocusableChild is None:
            self._cacheHasFocusableChild = False
            for child in self._children:
                if child._focusable or child._hasFocusableChild:
                    self._cacheHasFocusableChild = True
                    break
        return self._cacheHasFocusableChild

    def find(self, predicate):
        if predicate(self):
            return self
        for child in self._children:
            result = child.find(predicate)
            if result:
                return result

    def isLeafNode(self):
        if not self._children:
            return True

        # These types of objects may have children that we use as internal
        # implementation details, but we want to expose them as leaves to platform
        # accessibility APIs because screen readers might be confused if they find
        # any children.
        if self._isPlainTextField or self._isTextOnlyObject:
            return True
        # Roles whose children are only presentational according to the ARIA and
        # HTML5 Specs should be hidden from screen readers.
        # (Note that whilst ARIA buttons can have only presentational children, HTML5
        # buttons are allowed to have content.)
        if self._role in [
            'doc-cover',
            'graphics-symbol',
            'img',
            'Meter',
            'scrollbar',
            'slider',
            'separator',
            'progressbar',
        ]:
            return True

        # here and below: android heuristics
        if self._hasFocusableChild:
            return False
        if self._focusable and self._name:
            return True
        if self._role == 'heading' and self._name:
            return True
        return False

    @property
    def isControl(self):
        return self._role in [
            'button',
            'checkbox',
            'ColorWell',
            'combobox',
            'DisclosureTriangle',
            'listbox',
            'menu',
            'menubar',
            'menuitem',
            'menuitemcheckbox',
            'menuitemradio',
            'radio',
            'scrollbar',
            'searchbox',
            'slider',
            'spinbutton',
            'switch',
            'tab',
            'textbox',
            'tree',
        ]

    def isInteresting(self, insideControl: bool) -> bool:
        role = self._role
        if role == 'Ignored' or self._hidden:
            return False
        if self._focusable or self._richlyEditable:
            return True

        # If it's not focusable but has a control role, then it's interesting.
        if self.isControl:
            return True

        # A non focusable child of a control is not interesting
        if insideControl:
            return False
        return self.isLeafNode() and bool(self._name)

    def serialize(self):  # noqa C901
        properties: Dict[str, Union[str, float, bool]] = {}
        for property_ in self._payload.get('properties', []):
            properties[property_['name'].lower()] = property_['value'].get('value')
        if self._payload.get('name'):
            properties['name'] = self._payload['name']['value']
        if self._payload.get('value'):
            properties['value'] = self._payload['value']['value']
        if self._payload.get('description'):
            properties['description'] = self._payload['description']['value']

        node = {'role': self._role}
        userStringProperties = [
            'name',
            'value',
            'description',
            'keyshortcuts',
            'roledescription',
            'valuetext',
        ]
        for prop in userStringProperties:
            if prop not in properties:
                continue
            node[prop] = properties[prop]
        booleanProperties = [
            'disabled',
            'expanded',
            'focused',
            'modal',
            'multiline',
            'multiselectable',
            'readonly',
            'required',
            'selected',
        ]
        for prop in booleanProperties:
            # WebArea's treat focus differently than other nodes. They report whether their frame  has focus,
            # not whether focus is specifically on the root node.
            if prop == 'focused' and self._role == 'WebArea':
                continue
            value = properties.get(prop)
            if value:
                node[prop] = value

        tristateProperties = [
            'checked',
            'pressed',
        ]
        for prop in tristateProperties:
            if prop not in properties:
                continue
            value = properties.get(prop)
            if value != 'mixed':
                if value == 'true':
                    value = True
                else:
                    value = False
            node[prop] = value

        numericProperties = [
            'level',
            'valuemax',
            'valuemin',
        ]
        for prop in numericProperties:
            if prop not in properties:
                continue
            node[prop] = properties.get(prop)

        tokenProperties = [
            'autocomplete',
            'haspopup',
            'invalid',
            'orientation',
        ]
        for prop in tokenProperties:
            value = properties.get(prop)
            if not value or value == 'false':
                continue
            node[prop] = value
        return node

    @staticmethod
    def createTree(payloads: List[dict]) -> 'AXNode':
        """

        :param payloads: List of dictionaries of AXNode kwargs
        :return:
        """
        nodeById = {}
        for payload in payloads:
            nodeById[payload['nodeId']] = AXNode(payload)
        for node in nodeById.values():
            for childId in node._payload.get('childIds', []):
                node._children.append(nodeById[childId])
        return [*nodeById.values()][0]


def collectInterestingNodes(collection: Set[AXNode], node: AXNode, insideControl: bool):
    if node.isInteresting(insideControl):
        collection.add(node)
    if node.isLeafNode():
        return
    insideControl = insideControl or node.isControl
    for child in node._children:
        collectInterestingNodes(collection, child, insideControl)


def serializeTree(node: 'AXNode', whitelistedNodes: Set[AXNode] = None):
    children = []
    for child in node._children:
        children.extend(serializeTree(child, whitelistedNodes))
    if whitelistedNodes and node not in whitelistedNodes:
        return children
    serializedNode = node.serialize()
    if children:
        serializedNode['children'] = children
    return [serializedNode]



import glob
import shutil
import time
import zipfile
import os
import io
import shlex

from spython.main import Client # type: ignore
from spython.instance import Instance # type: ignore

from biolib import utils
from biolib.biolib_binary_format import ModuleOutput, ModuleInput
from biolib.biolib_singularity_client import BiolibSingularityClient
from biolib.biolib_logging import logger
from biolib.compute_node import utils as compute_node_utils
from biolib.compute_node.job_worker.executors.base_executor import BaseExecutor
from biolib.compute_node.job_worker.executors.types import StatusUpdate, LocalExecutorOptions
from biolib.compute_node.job_worker.mappings import Mappings, path_without_first_folder
from biolib.compute_node.job_worker.utils import ComputeProcessException
from biolib.compute_node.utils import SystemExceptionCodes
from biolib.typing_utils import List, Optional
from biolib.utils import get_absolute_container_image_uri


class SingularityExecutor(BaseExecutor):

    def __init__(self, options: LocalExecutorOptions):
        super().__init__(options)

        # Move this to shared class
        ##########
        if self._options['root_job_id'] == utils.RUN_DEV_JOB_ID:
            self._image_uri = self._options['module']['image_uri']
        else:
            self._image_uri = get_absolute_container_image_uri(
                base_url=self._options['biolib_base_url'],
                relative_image_uri=self._options['module']['image_uri'],
                job_is_federated=True if self._options['job'].get('federated_job_uuid') else False
            )

        self._send_system_exception = options['send_system_exception']
        self._send_stdout_and_stderr = options['send_stdout_and_stderr']

        federated_job_uuid: Optional[str] = self._options['job'].get('federated_job_uuid')
        if options['compute_node_info'] is not None and not federated_job_uuid:
            compute_node_public_id = options['compute_node_info']['public_id']
            compute_node_auth_token = options['compute_node_info']['auth_token']
            # Use "|" to separate the fields as this makes it easy for the ECR proxy to split
            ecr_proxy_auth_token = f'cloud-{compute_node_public_id}|{compute_node_auth_token}'
        else:
            ecr_proxy_auth_token = options['access_token']

        docker_auth_job_id = federated_job_uuid or self._options['job']['public_id']
        self._docker_auth_config = {'username': 'AWS', 'password': f'{ecr_proxy_auth_token},{docker_auth_job_id}'}
        self._random_docker_id = compute_node_utils.random_string(15)
        self._compute_process_dir = os.path.dirname(os.path.realpath(__file__))
        #########

        self._instance: Optional[Instance] = None
        self._input_files_dir = SingularityExecutor._create_dir_and_return_path(
            f'.biolib_singularity/biolib_input_{self._random_docker_id}'
        )
        self._runtime_files_dir = SingularityExecutor._create_dir_and_return_path(
            f'.biolib_singularity/biolib_runtime_{self._random_docker_id}'
        )
        self._output_files_dir = SingularityExecutor._create_dir_and_return_path(
            f'.biolib_singularity/biolib_output_{self._random_docker_id}'
        )
        self._pull_folder = SingularityExecutor._create_dir_and_return_path(
            '.biolib_singularity/images'
        )

        # The image file name is without the registry domain `containers.biolib.com/`
        self._image_file_name = f'{self._image_uri.split("/", 1)[1]}.sif'
        self._singularity_options: List[str] = []

        self._input_binds: List['str'] = []
        self._runtime_binds: List['str'] = []

        self._spython_client = BiolibSingularityClient.get_singularity_client()

    @staticmethod
    def _create_dir_and_return_path(path):
        os.makedirs(path, exist_ok=True)
        return path

    # Move this to shared class
    #############
    def execute_module(self, module_input_serialized: bytes) -> bytes:
        send_status_update = self._options['send_status_update']
        send_system_exception = self._options['send_system_exception']

        module_input = ModuleInput(module_input_serialized).deserialize()

        # TODO: fix these status updates such that they also make sense for run-dev
        send_status_update(StatusUpdate(progress=55, log_message='Pulling images...'))

        self._pull()

        send_status_update(StatusUpdate(progress=70, log_message='Computing...'))
        start_time = time.time()

        stdout, stderr, exit_code, mapped_output_files = self._execute_helper(module_input)

        try:
            module_output_serialized: bytes = ModuleOutput().serialize(stdout, stderr, exit_code, mapped_output_files)
            logger.debug(f'Compute time: {time.time() - start_time}')
            return module_output_serialized

        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_SERIALIZE_AND_SEND_MODULE_OUTPUT.value,
                send_system_exception
            ) from exception
        finally:
            try:
                self.cleanup()
            except Exception:  # pylint: disable=broad-except
                logger.error('SingularityExecutor failed to clean up container')
    #############

    def _pull(self):
        try:
            # Don't pull if SIF already exists
            if os.path.exists(self._image_file_name):
                return

            start_time = time.time()
            os.environ['SINGULARITY_DOCKER_USERNAME'] = self._docker_auth_config['username']
            os.environ['SINGULARITY_DOCKER_PASSWORD'] = self._docker_auth_config['password']
            self._spython_client.pull(
                image=f'docker://{self._image_uri}',
                pull_folder=self._pull_folder
            )
            logger.debug(f'Pulled image in: {time.time() - start_time}')

        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_PULL_DOCKER_IMAGE.value,
                self._send_system_exception,
                may_contain_user_data=False
            ) from exception

    def _execute_helper(self, module_input):
        runtime_files = {}
        if self._options['runtime_zip_bytes']:
            runtime_files = self.get_runtime_files(self._options['runtime_zip_bytes'])
            self._runtime_binds = SingularityExecutor.populate_bind_mount_dir(
                mappings=self._options['module']['source_files_mappings'],
                bind_dir=self._runtime_files_dir,
                arguments=module_input['arguments'],
                files=runtime_files
            )

        self._input_binds = SingularityExecutor.populate_bind_mount_dir(
            mappings=self._options['module']['input_files_mappings'],
            bind_dir=self._input_files_dir,
            arguments=module_input['arguments'],
            files=module_input['files']
        )

        try:
            # Set network to None to disable networking in container
            self._singularity_options = [
                '--net',
                '--network',
                'none',
                '--writable',
                *self._input_binds,
                *self._runtime_binds,
                '--bind',
                f'{self._output_files_dir}:/output',
            ]

            self._instance = self._spython_client.instance(
                image=f'{self._pull_folder}/{self._image_file_name}',
                options=self._singularity_options,
                start=True
            )

            logger.debug('Running Singularity container')
            result = Client.run(
                image=self._instance,
                args=shlex.split(self._options['module']['command']) + module_input['arguments'],
                contain=True,
                nv=False,  # TODO: Set to true when GPUs are needed
                stream=False,  # TODO: Set to true to test output streaming
                return_result=True
            )

            exit_code = result['return_code']
            # 137 is the error code from linux OOM killer (Should catch 90% of OOM errors)
            if exit_code == 137:
                raise ComputeProcessException(
                    MemoryError(),
                    SystemExceptionCodes.OUT_OF_MEMORY.value,
                    self._send_system_exception
                )

            logger.debug(f'Singularity container exited with code {exit_code}')

            full_stdout = result['message'].encode()
            full_stderr = b''  # TODO: Figure out a way to get STDERR from singularity

            mapped_output_files = self._get_output_files(arguments=module_input['arguments'])
            return full_stdout, full_stderr, exit_code, mapped_output_files

        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_RUN_COMPUTE_CONTAINER.value,
                self._send_system_exception
            ) from exception

    def cleanup(self):
        # Don't clean up if already in the process of doing so, or done doing so
        if self._is_cleaning_up:
            return
        else:
            self._is_cleaning_up = True

        bind_dir_time = time.time()
        for path_to_delete in [self._input_files_dir, self._runtime_files_dir, self._output_files_dir]:
            if os.path.exists(path_to_delete):
                shutil.rmtree(path_to_delete)
        logger.debug(f"Deleted bind dirs in: {time.time() - bind_dir_time}")

        instance_time = time.time()
        if self._instance:
            try:
                self._instance.stop(timeout=0)
            except Exception:  # pylint: disable=broad-except
                logger.debug('Instance could not be stopped or is already stopped')
        logger.debug(f"Deleted compute instance in: {time.time() - instance_time}")


    @staticmethod
    def populate_bind_mount_dir(mappings, bind_dir, arguments: List[str], files):
        mappings = Mappings(mappings, arguments)
        bind_mounts = []
        for path, data in files.items():
            # Skip the empty root entry that occurs in the runtime zip
            if path == '/':
                continue

            # Make all paths absolute
            if not path.startswith('/'):
                path = '/' + path

            mapped_file_names = mappings.get_mappings_for_path(path)
            for mapped_file_name in mapped_file_names:
                host_path = f'{bind_dir}{path}'
                if path.endswith('/'):
                    os.makedirs(host_path, exist_ok=True)
                else:
                    open(host_path, 'wb').write(data)
                bind_mounts.extend(['--bind', f'{host_path}:{mapped_file_name}'])
        return bind_mounts

    def get_runtime_files(self, runtime_zip_data, remove_root_folder=True):
        runtime_zip = zipfile.ZipFile(io.BytesIO(runtime_zip_data))
        runtime_files = {}
        for zip_file_name in runtime_zip.namelist():
            # Make paths absolute and remove root folder from path
            if remove_root_folder:
                file_path = '/' + path_without_first_folder(zip_file_name)
            else:
                file_path = '/' + zip_file_name

            runtime_files[file_path] = runtime_zip.read(zip_file_name)
        return runtime_files

    def _get_output_files(self, arguments: List[str]):
        module = self._options['module']
        mapped_output_files = {}

        try:
            # Write list of output files to retrieve
            files_to_get = [mapping['from_path'] for mapping in module['output_files_mappings']]
            with open(f'{self._output_files_dir}/filelist.txt', 'w') as file:
                for file_path in files_to_get:
                    file.write(f'{file_path}\n')

            bash_script = """
            #!/bin/sh
            cat /output/filelist.txt | while read line 
            do
                echo $line
                cp -r $line /output
            done
            """

            with open(f'{self._output_files_dir}/output_retriever.sh', 'w') as file:
                file.write(bash_script)

            Client.run(
                image=self._instance,
                args=['/bin/sh', '-c', 'chmod +x /output/output_retriever.sh && /output/output_retriever.sh'],
                options=self._singularity_options,
                contain=True,
            )

            output_mappings = Mappings(module['output_files_mappings'], arguments)
            for original_file_name in glob.iglob(self._output_files_dir + '**/**', recursive=True):
                # Remove the `output_files_dir` folder from file_name
                file_name = original_file_name.replace(self._output_files_dir, '')
                if file_name != '/':
                    mapped_file_names = output_mappings.get_mappings_for_path(file_name)
                    for mapped_file_name in mapped_file_names:
                        mapped_output_files[mapped_file_name] = open(original_file_name, 'rb').read()

        except Exception as exception:
            raise ComputeProcessException(
                exception,
                SystemExceptionCodes.FAILED_TO_RUN_COMPUTE_CONTAINER.value,
                self._send_system_exception
            ) from exception

        return mapped_output_files



from biolib.biolib_binary_format import BioLibBinaryFormatBasePackage
from biolib.biolib_binary_format.base_package import IndexableBuffer, InMemoryIndexableBuffer
from biolib.biolib_binary_format.common_types import FilesInfo
from biolib.biolib_binary_format.module_output import ModuleOutput
from biolib.typing_utils import TypedDict, List


class UnencryptedModuleOutputMetadata(TypedDict):
    version: int
    type: int
    stdout_length: int
    stderr_length: int
    files_info_length: int
    files_data_length: int


unencrypted_module_output_metadata_lengths = UnencryptedModuleOutputMetadata(
    # Note: order is important
    version=1,
    type=1,
    stdout_length=8,
    stderr_length=8,
    files_info_length=8,
    files_data_length=8,
)


class UnencryptedModuleOutputError(BaseException):
    pass


class UnencryptedModuleOutput(BioLibBinaryFormatBasePackage):
    _TYPE = 11
    _VERSION = 1

    _FILE_PATH_LENGTH_BYTES = 4
    _FILE_DATA_LENGTH_BYTES = 8

    def __init__(self, buffer: IndexableBuffer):
        super().__init__()
        self._buffer = buffer

    def _get_metadata(self) -> UnencryptedModuleOutputMetadata:
        if not self._buffer:
            raise UnencryptedModuleOutputError(
                'You must construct this class with a buffer to get its metadata'
            )

        metadata = {}
        for field_name, field_length in unencrypted_module_output_metadata_lengths.items():
            value = self._buffer.get_data_with_pointer_as_int(length_bytes=field_length)  # type: ignore
            if field_name == 'version' and value != UnencryptedModuleOutput._VERSION:
                raise Exception('Version does not match')

            if field_name == 'type' and value != UnencryptedModuleOutput._TYPE:
                raise Exception('Type does not match')

            metadata[field_name] = value

        return UnencryptedModuleOutputMetadata(**metadata)  # type: ignore

    def convert_to_serialized_module_output(self) -> bytes:
        if not self._buffer:
            raise UnencryptedModuleOutputError(
                'You must construct this class with a buffer to convert to module output'
            )

        metadata = self._get_metadata()

        exit_code = self._buffer.get_data_with_pointer_as_int(length_bytes=2)
        stdout = self._buffer.get_data_with_pointer(length_bytes=metadata['stdout_length'])
        stderr = self._buffer.get_data_with_pointer(length_bytes=metadata['stderr_length'])

        files_info_serialized = InMemoryIndexableBuffer(
            self._buffer.get_data_with_pointer(length_bytes=metadata['files_info_length'])
        )
        files_info: List[FilesInfo] = []
        while files_info_serialized.pointer <= len(files_info_serialized) - 1:  # subtract 1 as pointer is 0-indexed
            path_length = files_info_serialized.get_data_with_pointer_as_int(length_bytes=4)
            files_info.append({
                'path': files_info_serialized.get_data_with_pointer(length_bytes=path_length).decode(),
                'data_length': files_info_serialized.get_data_with_pointer_as_int(length_bytes=8)
            })

        files = {}
        for file_info in files_info:
            files[file_info['path']] = self._buffer.get_data_with_pointer(
                length_bytes=file_info['data_length']
            )

        return ModuleOutput().serialize(
            exit_code=exit_code,
            files=files,
            stderr=stderr,
            stdout=stdout,
        )

    @staticmethod
    def create_from_serialized_module_output(module_output_serialized: bytes) -> bytes:
        module_output_dict = ModuleOutput(module_output_serialized).deserialize()

        bbf_data = bytearray()
        bbf_data.extend(UnencryptedModuleOutput._VERSION.to_bytes(1, 'big'))
        bbf_data.extend(UnencryptedModuleOutput._TYPE.to_bytes(1, 'big'))

        # Length of stdout and stderr
        bbf_data.extend(len(module_output_dict['stdout']).to_bytes(8, 'big'))
        bbf_data.extend(len(module_output_dict['stderr']).to_bytes(8, 'big'))

        files_info = bytearray()
        files_data = bytearray()
        for path, data in module_output_dict['files'].items():
            encoded_path = path.encode()
            files_info.extend(len(encoded_path).to_bytes(4, 'big'))
            files_info.extend(encoded_path)
            files_info.extend(len(data).to_bytes(8, 'big'))

            files_data.extend(data)

        bbf_data.extend(len(files_info).to_bytes(8, 'big'))
        bbf_data.extend(len(files_data).to_bytes(8, 'big'))

        bbf_data.extend(module_output_dict['exit_code'].to_bytes(2, 'big'))
        bbf_data.extend(module_output_dict['stdout'])
        bbf_data.extend(module_output_dict['stderr'])
        bbf_data.extend(files_info)
        bbf_data.extend(files_data)

        return bbf_data
